#!/usr/bin/python2.2

from Tkinter import *
import os, sys
import math
import bisect

class Cord(object):
  
  def __init__(self, x, y):
    self.x = x
    self.y = y

class ColorGenerator(list):

  def __getitem__(self, i):
    try:
      return list.__getitem__(self, i)
    except IndexError:
      # generate new hsv-color
      b = 1 # base of 2^x
      while i >= b: b *= 2
      h = (360.0/b)%360 + 720.0/b*(i-b/2)
      s = 1.0 / 1.1**(b-1); v = s/2 + 0.5
      r,g,b = hsv_to_rgb(h,s,v)
      rgb = '#%02x%02x%02x' % (r*255,g*255,b*255)
      self.append(rgb)
      return self[i]
   
def hsv_to_rgb(h,s,v):
  # algorithm found at: http://www.cs.rit.edu/~ncs/color/t_convert.html

  if(s == 0) : # achromatic (grey)
    return [v,v,v]
  else:
    h = h/60.0
    i = int(h) # sector 0 to 5
    f = h - i  # factorial part of h
    p = v * ( 1 - s )
    q = v * ( 1 - s * f )
    t = v * ( 1 - s * ( 1 - f ) )
    return [(v,t,p), (q,v,p), (p,v,t), (p,q,v), (t,p,v), (v,p,q)][i]

class MultiselectListbox(Frame):
    def __init__(self, master, lists, colorlist=None, command=None):
        Frame.__init__(self, master)
        self.lists = []
        self.colorlist = colorlist
        self._color_index = []
        self._selected = {} # n: color_index
        self.command = command
        for l,w in lists:
            frame = Frame(self); frame.pack(side=LEFT, expand=YES, fill=BOTH)
            Label(frame, text=l, borderwidth=1, relief=RAISED).pack(fill=X)
            lb = Listbox(frame, width=w, borderwidth=0, selectborderwidth=0,
                         relief=FLAT, exportselection=FALSE, selectbackground='white')
            lb.pack(expand=YES, fill=BOTH)
            self.lists.append(lb)
            lb.bind('<Button-1>', lambda e, s=self: s._select(e.y))
            lb.bind('<B1-Motion>', lambda e: 'break')
            lb.bind('<Leave>', lambda e: 'break')
            lb.bind('<B2-Motion>', lambda e, s=self: s._b2motion(e.x, e.y))
            lb.bind('<Button-2>', lambda e, s=self: s._button2(e.x, e.y))
        frame = Frame(self); frame.pack(side=LEFT, fill=Y)
        Label(frame, borderwidth=1, relief=RAISED).pack(fill=X)
        sb = Scrollbar(frame, orient=VERTICAL, command=self._scroll)
        sb.pack(expand=YES, fill=Y)
        self.lists[0]['yscrollcommand']=sb.set

    def _select(self, y):
        row = self.lists[0].nearest(y)
        if row in self._selected: # selected
          bisect.insort(self._color_index, self._selected[row])
          del self._selected[row]
          color = ''
          if self.command:
            self.command(row, None)
        else: # not selected
          try:
            if self._color_index:
              color_index = self._color_index.pop(0)
            else:
              color_index = len(self._selected)
            color = self.colorlist[color_index]
          except IndexError:
            color = '#999'
          self._selected[row] = color_index
          if self.command:
            self.command(row, color)
        for l in self.lists:
          l.itemconfig(row, bg=color)
        return 'break'

    def _button2(self, x, y):
        for l in self.lists: l.scan_mark(x, y)
        return 'break'

    def _b2motion(self, x, y):
        for l in self.lists: l.scan_dragto(x, y)
        return 'break'

    def _scroll(self, *args):
        for l in self.lists:
            apply(l.yview, args)

    def insert(self, index, *elements):
        for e in elements:
            i = 0
            for l in self.lists:
                l.insert(index, e[i])
                i = i + 1

class TeamList(MultiselectListbox):

  def __init__(self, parent, command=None):
    MultiselectListbox.__init__(self, parent, (('Team',20), ('Res',5)), colorlist = ColorGenerator(), command=command)

START, FINISH = ('start','finish')

class Map(Frame):

  def __init__(self, parent):
    Frame.__init__(self, parent)
    self.grid_rowconfigure(0, weight=1)
    self.grid_columnconfigure(0, weight=1)

    self.xscrollbar = Scrollbar(self, orient=HORIZONTAL)
    self.xscrollbar.grid(row=1, column=0, sticky=E+W)
    self.yscrollbar = Scrollbar(self, orient=VERTICAL)
    self.yscrollbar.grid(row=0, column=1, sticky=N+S)

    self.canvas = Canvas(self, xscrollcommand=self.xscrollbar.set,
                               yscrollcommand=self.yscrollbar.set)
    self.canvas.grid(row=0, column=0, sticky=N+S+E+W)
    self.xscrollbar.config(command=self.canvas.xview)
    self.yscrollbar.config(command=self.canvas.yview)
    
class Time(Scale):

  limits = [3.0, 8.0, 16.0, 24.0]

  def __init__(self, parent, (timemin, timemax), limit=None, change=None):
    timemin = math.floor(timemin)
    timemax = math.ceil(timemax)
    Scale.__init__(self, parent, orient=HORIZONTAL, showvalue=1, from_=timemin, to=timemax, resolution=0.05,
                   command=self.handler)
    self.change = change # callback when changed 
    self.set(limit)
    self.current = self.get()

  def handler(self, event):
    if self.current != self.get():
      self.current = self.get()
      if self.change:
        self.change(self.current)

class ControlPoint(object):

  def __repr__(self):
    return 'P '+self.name

  def __init__(self, name, cord, flags):
    if name[:2] == 'KP':
      name = name[2:]
    self.name = name
    self.cord = cord
    self.flags = flags

  COLOR="#f0f"
  RAD=10
  FINISH_RAD=6

  def special(self):
    return {'0': START, '6':FINISH}.get(self.flags)

  def draw(self, canvas, scale):
    x,y = self.cord.x * scale / 100, self.cord.y * scale / 100
    x0, y0, x1, y1 = x-self.RAD, y-self.RAD, x+self.RAD, y+self.RAD
    try:
      canvas.coords(self._line, x0, y0, x1, y1)
    except AttributeError:
      self._line = canvas.create_oval(x0, y0, x1, y1,
                                      outline=self.COLOR, width=2)
    if self.name.isdigit():
      x0, y0 = x+self.RAD, y-self.RAD/2
      try:
        canvas.coords(self._text, x0, y0)
      except AttributeError:
        self._text = canvas.create_text(x0, y0,
                                        anchor=SW, text=str(self.name), fill=self.COLOR)
    if self.flags == '6':
      x0, y0, x1, y1 = x-self.FINISH_RAD, y-self.FINISH_RAD, x+self.FINISH_RAD, y+self.FINISH_RAD
      try:
        canvas.coords(self._line2, x0, y0, x1, y1)
      except AttributeError:
        self._line2 = canvas.create_oval(x0, y0, x1, y1,
                                         outline=self.COLOR, width=2)

class Team(object):

  def __init__(self, name, total, route=[], id=0):
    self.id = id
    self.name = name
    self.total = total
    self.route = route # [('start', time), ('point', time), ...]
    self.lengths = None
    self.color = None

  def __cmp__(self, other):
    return cmp(self.id, other.id)

  def bind(self, legmap, color=None):
    self.color = color
    self.lengths = {}

  def unbind(self, legmap):
    for leg in self.lengths.iterkeys():
      legmap.get(leg).unbind(self)
    self.color = None
    self.lengths = None

  def settime(self, time, legmap):
    lengths = {}
    p0, t0 = self.route[0]
    for p,t in self.route[1:]:
      if time <= t0: break
      leg = (p0,p)
      length = (time >= t) and 1.0 or ((time-t0) / (t-t0))
      try:
        assert length == self.lengths[leg]
        del self.lengths[leg]
      except KeyError: # didn't exist
        legmap.get(leg).bind(self)
      except AssertionError: # changed
        legmap.get(leg).update()
        del self.lengths[leg]
      lengths[leg] = length
      p0, t0 = p, t
    for leg in self.lengths.keys():
      legmap.get(leg).unbind(self)
    self.lengths = lengths

ascend = lambda a,b: (a<=b) and (a,b) or (b,a)

class Leg(list):

  WIDTH = 3

  def __init__(self, id, start, end, _update):
    self.id = id
    self.start = start
    self.end = end
    self.length = math.sqrt((end.x-start.x)**2.0 + (end.y-start.y)**2.0)
    self.vector = Cord((end.x-start.x)/self.length, (end.y-start.y)/self.length)
    self._update = _update # update queue function
    self._lines = {}
  
  def bind(self, team):
    bisect.insort(self, team)
    self._update(self.id)

  def unbind(self, team):
    self.remove(team)
    self._update(self.id)

  def update(self):
    self._update(self.id)

  def draw(self, canvas, width, scale):
    _lines = {}
    pos = (width-1)/2.0 - (len(self)-1)
    X0 = self.start.x*scale/100 + self.vector.x*ControlPoint.RAD
    Y0 = self.start.y*scale/100 + self.vector.y*ControlPoint.RAD
    X1 = self.end.x*scale/100
    Y1 = self.end.y*scale/100
    hidden_start = ControlPoint.RAD / self.length*scale/100 # 0 .. 1.0
    for team in self:
      length = team.lengths[self.id]
      if length > hidden_start:
        x0 = X0 - self.vector.y*Leg.WIDTH*pos
        y0 = Y0 + self.vector.x*Leg.WIDTH*pos
        if length == 1.0:
          x1 = X1 - self.vector.y*Leg.WIDTH*pos
          y1 = Y1 + self.vector.x*Leg.WIDTH*pos
        else:
          x1 = X0*(1.0-length) + X1*length - self.vector.y*Leg.WIDTH*pos
          y1 = Y0*(1.0-length) + Y1*length + self.vector.x*Leg.WIDTH*pos
        try: 
          line = self._lines[team]
          canvas.coords(line, x0,y0,x1,y1)
        except KeyError:
          line = canvas.create_line(x0,y0,x1,y1, fill=team.color, width=Leg.WIDTH)
        else:
          del self._lines[team]
        _lines[team] = line
      pos += 1
    for line in self._lines.values():
      canvas.delete(line)    
    self._lines = _lines

class Legmap(dict):
  
  def __init__(self, controlpoints):
    self.controlpoints = controlpoints
    self._update = {}

  def get(self, leg):
    a,b = leg
    p0 = self.controlpoints[a]
    p1 = self.controlpoints[b]
    return self.setdefault(leg, Leg(leg, p0.cord, p1.cord, self.update))
  
  def draw(self, canvas):
    while self._update:
      (a,b),value = self._update.popitem()
      leg, leg2 = self.get((a,b)), self.get((b,a))
      try:
        del self._update[(b,a)]
      except KeyError:
        pass
      width = len(leg) + len(leg2)
      leg.draw(canvas, width, self.scale)
      leg2.draw(canvas, width, self.scale)

  def redraw(self, scale, canvas):
    self.scale = scale
    for leg in self.keys():
      self._update[leg] = 1
    self.draw(canvas)

  def update(self, leg):
    self._update[leg] = 1

minx = lambda *tup: min([x for x in tup if x != None])
  
class Prog(Frame):
   
  def __init__(self, parent, radafile):
    Frame.__init__(self, parent)

    self.readfiles(radafile)

    self.grid_rowconfigure(2, weight=1)
    self.grid_columnconfigure(1, weight=1)

    self.scale = IntVar()
    self.scale.set(100)

    self.menubar = Menu(self)

    menu1 = Menu(self.menubar, tearoff=0)
    menu1.add_command(label='Exit', command=parent.quit)
    self.menubar.add_cascade(label='File', menu=menu1)
 
    menu2 = Menu(self.menubar, tearoff=0)
    for s in (100,50,25):
      menu2.add_radiobutton(label=str(s)+'%', variable=self.scale, value=s, command=self.update_scale)
    self.menubar.add_cascade(label='Map', menu=menu2)

    parent.config(menu=self.menubar)

    self.teamlist = TeamList(self, command=self.update_selected)
    for team in self.teams:
      self.teamlist.insert(END, (team.name, team.total))
    self.teamlist.grid(row=1, column=0, rowspan=2, sticky=N+S)

    if not self.timelimit:
      try:
        self.timelimit = Time.limits[bisect.bisect_right(Time.limits, self.timerange[1])-1]
      except IndexError:
        self.timelimit = 0.0

    self.time = Time(self, self.timerange, limit=self.timelimit, change=self.update_time)
    self.time.grid(row=1, column=1, sticky=W+E)
    
    self.map = Map(self)
    self.map.grid(row=2, column=1, sticky=N+W+S+E)
    self.canvas = self.map.canvas

    self.legmap = Legmap(self.controlpoints)

    self.update_scale()
    self.update_time(self.timelimit)
    
  def readfiles(self, radafile):
    self.teams = []
    self.selected = []
    self.controlpoints = {}
    self.timerange = (0.0, 0.0)
    self.timelimit = None

    radafile = os.path.expanduser(radafile)
    resfile = os.path.splitext(radafile)[0] + '.aeg'
 
    for line in open(radafile).readlines():
      while line[-1:] in '\n\r': line = line[:-1]
      if line.startswith('map='):
        self.mapfile = line.split('=',1)[1]
      elif line.startswith('dpi='):
        self.dpi = float(line.split('=',1)[1])
      elif line.startswith('mastap='):
        self.mapscale = float(line.split('=',1)[1])
      elif line.startswith('limit='):
        self.timelimit = float(line.split('=',1)[1])
      elif line.split('=',1)[0].isdigit():
        x,y,flags,name = line.split('=',1)[1].split('|',3)
        x,y = map(int, (x,y))
        controlpoint = ControlPoint(name,Cord(x,y),flags)
        self.controlpoints[controlpoint.name] = controlpoint
        if controlpoint.special():
          self.controlpoints[controlpoint.special()] = controlpoint

    self.original_image = PhotoImage(file = self.mapfile)

    for line in open(resfile).readlines():
      self.teams.append(self._parseteam(line, count=len(self.teams)))

  def _parseteam(self, line, count):
    parts = line.rstrip('\n\r\t').split('\t')
    total, name = parts[0:2]
    route = map(self._parsepoints, parts[2:])
    return Team(name, total, route, id=count)

  def _parsepoints(self, txt):
    p,t = txt.split('|')
    h,m,s = map(int, t.split(':'))
    hours = h + m/60.0 + s /3600.0
    self.timerange = (minx(self.timerange[0], hours), max(self.timerange[1], hours))
    return self.controlpoints[p].name, hours

  def update_scale(self):
    scale = self.scale.get()
 
    self._image = self.original_image.subsample(100/scale)
    try:
      self.canvas.itemconfig(self._imageid, image=self._image)
    except AttributeError:
      self._imageid = self.canvas.create_image(0,0, image=self._image, anchor=NW)
    self.canvas.config(scrollregion=(0,0, self._image.width(), self._image.height()))

    for controlpoint in self.controlpoints.values():
      controlpoint.draw(self.canvas, scale)

    self.legmap.redraw(scale, self.canvas)

  def update_time(self, time):
    self.time = time
    for team in self.selected:
      team.settime(time, self.legmap)
    self.legmap.draw(self.canvas)

  def update_selected(self, i, color):
    print 'update',i,'to color',color
    team = self.teams[i]
    if color:
      team.bind(self.legmap, color=color)
      team.settime(self.time, self.legmap)
      self.selected.append(team)
    else: # removed
      self.teams[i].unbind(self.legmap)
      self.selected.remove(team)
    self.legmap.draw(self.canvas)
  
assert len(sys.argv) == 2, 'usage: rogain rada-file'
root = Tk()
prog = Prog(root, sys.argv[1])
prog.pack(side=LEFT, fill=BOTH, expand=YES)
root.mainloop()
